@@ -7,8 +7,6 @@ tonic::include_proto!("indradb");
 pub use bulk_insert_item::Item as BulkInsertItemVariant;
 pub use edge_query::Query as EdgeQueryVariant;
 pub use indra_db_client::IndraDbClient as ProtoClient;
-pub use transaction_request::Request as TransactionRequestVariant;
-pub use transaction_response::Response as TransactionResponseVariant;
 pub use vertex_query::Query as VertexQueryVariant;
 
 mod converters;
@@ -17,14 +15,12 @@ pub use converters::*;
 #[cfg(feature = "client")]
 mod client;
 #[cfg(feature = "client")]
-pub use client::{Client, ClientError, Transaction};
+pub use client::{Client, ClientError};
 
 #[cfg(feature = "server")]
 mod server;
 #[cfg(feature = "server")]
-pub use server::run as run_server;
-#[cfg(feature = "server")]
-pub use server::Server;
+pub use server::{run as run_server, run_with_plugins as run_server_with_plugins, Server};
 
 #[cfg(feature = "test-suite")]
-mod tests;
+pub mod tests;
