@@ -16,20 +16,20 @@ use tonic::{Request, Response, Status, Streaming};
 
 const CHANNEL_CAPACITY: usize = 100;
 
-async fn send<IT, PT>(tx: mpsc::Sender<Result<PT, Status>>, result: Result<Vec<IT>, indradb::Error>)
+fn send<IT, PT>(tx: mpsc::Sender<Result<PT, Status>>, result: Result<Vec<IT>, indradb::Error>)
 where
     IT: Into<PT>,
 {
     match map_indradb_result(result) {
         Ok(values) => {
             for value in values {
-                if let Err(err) = tx.send(Ok(value.into())).await {
+                if let Err(err) = tx.blocking_send(Ok(value.into())) {
                     eprintln!("could not send message to client: {}", err);
                 }
             }
         }
         Err(err) => {
-            if let Err(err) = tx.send(Err(err)).await {
+            if let Err(err) = tx.blocking_send(Err(err)) {
                 eprintln!("could not send message to client: {}", err);
             }
         }
@@ -44,6 +44,11 @@ fn map_conversion_result<T>(res: Result<T, crate::ConversionError>) -> Result<T,
     res.map_err(|err| Status::invalid_argument(format!("{}", err)))
 }
 
+fn map_jh_indra_result<T>(res: Result<Result<T, indradb::Error>, tokio::task::JoinError>) -> Result<T, Status> {
+    let jh_res = res.map_err(|err| Status::internal(format!("{}", err)))?;
+    map_indradb_result(jh_res)
+}
+
 /// An error that occurred while initializing the server with plugins enabled.
 #[derive(Debug)]
 pub enum InitError {
@@ -205,13 +210,17 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
     }
 
     async fn sync(&self, _: Request<()>) -> Result<Response<()>, Status> {
-        map_indradb_result(self.datastore.sync())?;
+        let datastore = self.datastore.clone();
+
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.sync()).await)?;
         Ok(Response::new(()))
     }
 
     async fn create_vertex(&self, request: Request<crate::Vertex>) -> Result<Response<crate::CreateResponse>, Status> {
+        let datastore = self.datastore.clone();
+
         let vertex = map_conversion_result(request.into_inner().try_into())?;
-        let res = map_indradb_result(self.datastore.create_vertex(&vertex))?;
+        let res = map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.create_vertex(&vertex)).await)?;
         Ok(Response::new(crate::CreateResponse { created: res }))
     }
 
@@ -219,8 +228,10 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         &self,
         request: Request<crate::Identifier>,
     ) -> Result<Response<crate::Uuid>, Status> {
+        let datastore = self.datastore.clone();
+
         let t = map_conversion_result(request.into_inner().try_into())?;
-        let res = map_indradb_result(self.datastore.create_vertex_from_type(t))?;
+        let res = map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.create_vertex_from_type(t)).await)?;
         Ok(Response::new(res.into()))
     }
 
@@ -230,45 +241,57 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         request: Request<crate::VertexQuery>,
     ) -> Result<Response<Self::GetVerticesStream>, Status> {
         let datastore = self.datastore.clone();
+
         let q: indradb::VertexQuery = map_conversion_result(request.into_inner().try_into())?;
         let (tx, rx) = mpsc::channel(CHANNEL_CAPACITY);
-        tokio::spawn(async move {
-            send(tx, datastore.get_vertices(q)).await;
+        tokio::task::spawn_blocking(move || {
+            let res = datastore.get_vertices(q);
+            send(tx, res)
         });
         Ok(Response::new(Box::pin(ReceiverStream::new(rx))))
     }
 
     async fn delete_vertices(&self, request: Request<crate::VertexQuery>) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let q: indradb::VertexQuery = map_conversion_result(request.into_inner().try_into())?;
-        map_indradb_result(self.datastore.delete_vertices(q))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.delete_vertices(q)).await)?;
         Ok(Response::new(()))
     }
 
     async fn get_vertex_count(&self, _: Request<()>) -> Result<Response<crate::CountResponse>, Status> {
-        let res = map_indradb_result(self.datastore.get_vertex_count())?;
+        let datastore = self.datastore.clone();
+
+        let res = map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.get_vertex_count()).await)?;
         Ok(Response::new(crate::CountResponse { count: res }))
     }
 
     async fn create_edge(&self, request: Request<crate::EdgeKey>) -> Result<Response<crate::CreateResponse>, Status> {
+        let datastore = self.datastore.clone();
+
         let key = map_conversion_result(request.into_inner().try_into())?;
-        let res = map_indradb_result(self.datastore.create_edge(&key))?;
+        let res = map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.create_edge(&key)).await)?;
         Ok(Response::new(crate::CreateResponse { created: res }))
     }
 
     type GetEdgesStream = Pin<Box<dyn Stream<Item = Result<crate::Edge, Status>> + Send + Sync + 'static>>;
     async fn get_edges(&self, request: Request<crate::EdgeQuery>) -> Result<Response<Self::GetEdgesStream>, Status> {
         let datastore = self.datastore.clone();
+
         let q: indradb::EdgeQuery = map_conversion_result(request.into_inner().try_into())?;
         let (tx, rx) = mpsc::channel(CHANNEL_CAPACITY);
-        tokio::spawn(async move {
-            send(tx, datastore.get_edges(q)).await;
+        tokio::task::spawn_blocking(move || {
+            let res = datastore.get_edges(q);
+            send(tx, res);
         });
         Ok(Response::new(Box::pin(ReceiverStream::new(rx))))
     }
 
     async fn delete_edges(&self, request: Request<crate::EdgeQuery>) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let q: indradb::EdgeQuery = map_conversion_result(request.into_inner().try_into())?;
-        map_indradb_result(self.datastore.delete_edges(q))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.delete_edges(q)).await)?;
         Ok(Response::new(()))
     }
 
@@ -276,8 +299,12 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         &self,
         request: Request<crate::GetEdgeCountRequest>,
     ) -> Result<Response<crate::CountResponse>, Status> {
+        let datastore = self.datastore.clone();
+
         let (id, t, direction) = map_conversion_result(request.into_inner().try_into())?;
-        let res = map_indradb_result(self.datastore.get_edge_count(id, t.as_ref(), direction))?;
+        let res = map_jh_indra_result(
+            tokio::task::spawn_blocking(move || datastore.get_edge_count(id, t.as_ref(), direction)).await,
+        )?;
         Ok(Response::new(crate::CountResponse { count: res }))
     }
 
@@ -288,10 +315,12 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         request: Request<crate::VertexPropertyQuery>,
     ) -> Result<Response<Self::GetVertexPropertiesStream>, Status> {
         let datastore = self.datastore.clone();
+
         let q = map_conversion_result(request.into_inner().try_into())?;
         let (tx, rx) = mpsc::channel(CHANNEL_CAPACITY);
-        tokio::spawn(async move {
-            send(tx, datastore.get_vertex_properties(q)).await;
+        tokio::task::spawn_blocking(move || {
+            let res = datastore.get_vertex_properties(q);
+            send(tx, res);
         });
         Ok(Response::new(Box::pin(ReceiverStream::new(rx))))
     }
@@ -303,11 +332,14 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         request: Request<crate::VertexQuery>,
     ) -> Result<Response<Self::GetAllVertexPropertiesStream>, Status> {
         let datastore = self.datastore.clone();
+
         let q: indradb::VertexQuery = map_conversion_result(request.into_inner().try_into())?;
         let (tx, rx) = mpsc::channel(CHANNEL_CAPACITY);
-        tokio::spawn(async move {
-            send(tx, datastore.get_all_vertex_properties(q)).await;
+        tokio::task::spawn_blocking(move || {
+            let res = datastore.get_all_vertex_properties(q);
+            send(tx, res);
         });
+
         Ok(Response::new(Box::pin(ReceiverStream::new(rx))))
     }
 
@@ -315,8 +347,10 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         &self,
         request: Request<crate::SetVertexPropertiesRequest>,
     ) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let (q, value) = map_conversion_result(request.into_inner().try_into())?;
-        map_indradb_result(self.datastore.set_vertex_properties(q, value))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.set_vertex_properties(q, value)).await)?;
         Ok(Response::new(()))
     }
 
@@ -324,8 +358,10 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         &self,
         request: Request<crate::VertexPropertyQuery>,
     ) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let q = map_conversion_result(request.into_inner().try_into())?;
-        map_indradb_result(self.datastore.delete_vertex_properties(q))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.delete_vertex_properties(q)).await)?;
         Ok(Response::new(()))
     }
 
@@ -336,10 +372,12 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         request: Request<crate::EdgePropertyQuery>,
     ) -> Result<Response<Self::GetEdgePropertiesStream>, Status> {
         let datastore = self.datastore.clone();
+
         let q: indradb::EdgePropertyQuery = map_conversion_result(request.into_inner().try_into())?;
         let (tx, rx) = mpsc::channel(CHANNEL_CAPACITY);
-        tokio::spawn(async move {
-            send(tx, datastore.get_edge_properties(q)).await;
+        tokio::task::spawn_blocking(move || {
+            let res = datastore.get_edge_properties(q);
+            send(tx, res);
         });
         Ok(Response::new(Box::pin(ReceiverStream::new(rx))))
     }
@@ -351,10 +389,12 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         request: Request<crate::EdgeQuery>,
     ) -> Result<Response<Self::GetAllEdgePropertiesStream>, Status> {
         let datastore = self.datastore.clone();
+
         let q: indradb::EdgeQuery = map_conversion_result(request.into_inner().try_into())?;
         let (tx, rx) = mpsc::channel(CHANNEL_CAPACITY);
-        tokio::spawn(async move {
-            send(tx, datastore.get_all_edge_properties(q)).await;
+        tokio::task::spawn_blocking(move || {
+            let res = datastore.get_all_edge_properties(q);
+            send(tx, res);
         });
         Ok(Response::new(Box::pin(ReceiverStream::new(rx))))
     }
@@ -363,18 +403,24 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
         &self,
         request: Request<crate::SetEdgePropertiesRequest>,
     ) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let (q, value) = map_conversion_result(request.into_inner().try_into())?;
-        map_indradb_result(self.datastore.set_edge_properties(q, value))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.set_edge_properties(q, value)).await)?;
         Ok(Response::new(()))
     }
 
     async fn delete_edge_properties(&self, request: Request<crate::EdgePropertyQuery>) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let q = map_conversion_result(request.into_inner().try_into())?;
-        map_indradb_result(self.datastore.delete_edge_properties(q))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.delete_edge_properties(q)).await)?;
         Ok(Response::new(()))
     }
 
     async fn bulk_insert(&self, request: Request<Streaming<crate::BulkInsertItem>>) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let items = {
             let mut stream = request.into_inner();
             let (lower_bound_stream_size, _) = stream.size_hint();
@@ -386,14 +432,15 @@ impl<D: indradb::Datastore + Send + Sync + 'static> crate::indra_db_server::Indr
             items
         };
 
-        let datastore = self.datastore.clone();
-        map_indradb_result(datastore.bulk_insert(items))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.bulk_insert(items)).await)?;
         Ok(Response::new(()))
     }
 
     async fn index_property(&self, request: Request<crate::IndexPropertyRequest>) -> Result<Response<()>, Status> {
+        let datastore = self.datastore.clone();
+
         let name: indradb::Identifier = map_conversion_result(request.into_inner().try_into())?;
-        map_indradb_result(self.datastore.clone().index_property(name))?;
+        map_jh_indra_result(tokio::task::spawn_blocking(move || datastore.index_property(name)).await)?;
         Ok(Response::new(()))
     }
 
